/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.voxelisation.als;

import fr.amap.amapvox.als.LasHeader;
import fr.amap.amapvox.als.LasPoint;
import fr.amap.amapvox.als.las.LasReader;
import fr.amap.amapvox.als.las.PointDataRecordFormat;
import fr.amap.amapvox.als.laz.LazExtraction;
import fr.amap.commons.util.io.file.FileManager;
import fr.amap.commons.math.matrix.Mat4D;
import fr.amap.commons.math.point.Point3D;
import fr.amap.commons.math.vector.Vec3D;
import fr.amap.commons.math.vector.Vec4D;
import fr.amap.commons.util.Process;
import fr.amap.commons.util.io.file.CSVFile;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.vecmath.Vector3d;
import fr.amap.commons.util.Cancellable;
import fr.amap.commons.util.IterableWithException;
import fr.amap.commons.util.IteratorWithException;
import fr.amap.lidar.amapvox.shot.Shot;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import javax.vecmath.Point3d;
import org.apache.log4j.Logger;

/**
 * This class merge trajectory file with point file (LAS, LAZ)
 *
 * @author Julien Heurtebize (julienhtbe@gmail.com)
 * @author Philippe Verley
 */
public class PointsToShot extends Process implements IterableWithException<Shot>, Cancellable {

    // LAS points
    private final File inputFile;
    private List<LasPoint> lasPoints;

    // trajectory
    private final CSVFile trajectoryFile;
    private List<TrajectoryPoint> trajectory;

    // VOP matrix
    private final Mat4D vopMatrix;

    private final double EPSILON = 0.01d;
    private final static float RATIO_REFLECTANCE_VEGETATION_SOL = 0.4f;
    private final static int SHOT_BUFFER = 100;

    // whether the task is cancelled
    private boolean cancelled;

    // logger
    private final static Logger LOGGER = Logger.getLogger(PointsToShot.class);

    public PointsToShot(CSVFile trajectoryFile, File inputFile, Mat4D vopMatrix) {

        this.trajectoryFile = trajectoryFile;
        this.vopMatrix = vopMatrix;
        this.inputFile = inputFile;
    }

    public void init() throws Exception {
        
        // read LAS / LAZ points
        lasPoints = readALSPoints(inputFile);

        // read trajectory file
        trajectory = readTrajectory(trajectoryFile);

        // reduce trajectory points to LAS points time range
        int imin = nearestTrajectoryPoint(lasPoints.get(0).t, Search.MIN);
        int imax = nearestTrajectoryPoint(lasPoints.get(lasPoints.size() - 1).t, Search.MAX);
        if (imin < 0 || imax < 0) {
            StringBuilder sb = new StringBuilder();
            sb.append("LAS points time span wider than trajectory file time span:\n");
            sb.append("LAS points time span [").append(lasPoints.get(0).t);
            sb.append(" ").append(lasPoints.get(lasPoints.size() - 1).t).append("]\n");
            sb.append("Trajectory time span [").append(trajectory.get(0).t);
            sb.append(" ").append(trajectory.get(trajectory.size() - 1).t).append("]");
            throw new IndexOutOfBoundsException(sb.toString());
        }
        trajectory = trajectory.subList(imin, imax + 1);
    }

    private List<LasPoint> readALSPoints(final File file) throws Exception {

        // read ALS points
        List<LasPoint> points = new ArrayList();
        LasHeader header;

        int count = 0;
        long nPoint;
        switch (FileManager.getExtension(file)) {
            case ".las":
                LOGGER.info("Reading LAS file " + file.getName());
                // open LAS reader
                LasReader lasReader = new LasReader();
                lasReader.open(file);
                // read LAS header
                header = lasReader.getHeader();
                // number of LAS points
                nPoint = header.getNumberOfPointrecords();
                // loop over LAS points
                for (PointDataRecordFormat p : lasReader) {
                    fireProgress("Reading LAS file " + file.getName(), count++, nPoint);
                    if (isCancelled()) {
                        return null;
                    }
                    Vector3d location = new Vector3d((p.getX() * header.getxScaleFactor()) + header.getxOffset(), (p.getY() * header.getyScaleFactor()) + header.getyOffset(), (p.getZ() * header.getzScaleFactor()) + header.getzOffset());
                    LasPoint point = new LasPoint(location.x, location.y, location.z, p.getReturnNumber(), p.getNumberOfReturns(), p.getIntensity(), p.getClassification(), p.getGpsTime());
                    points.add(point);
                }
                break;
            case ".laz":
                LOGGER.info("Reading LAZ file " + file.getName());
                // open LAZ reader
                LazExtraction laz = new LazExtraction();
                laz.openLazFile(file);
                // read LAZ header
                header = laz.getHeader();
                // number of LAS points
                nPoint = header.getNumberOfPointrecords();
                // loop over LAS points
                for (LasPoint p : laz) {
                    fireProgress("Reading LAZ file " + file.getName(), count++, nPoint);
                    if (isCancelled()) {
                        return null;
                    }
                    p.x = (p.x * header.getxScaleFactor()) + header.getxOffset();
                    p.y = (p.y * header.getyScaleFactor()) + header.getyOffset();
                    p.z = (p.z * header.getzScaleFactor()) + header.getzOffset();
                    points.add(p);
                }
                laz.close();
                break;
        }

        // sort LAS point by time stamp
        LOGGER.info("Sorting LAS points");
        fireProgress("Sorting LAS points", 50, 100);
        Collections.sort(points, (LasPoint lp1, LasPoint lp2) -> Double.compare(lp1.t, lp2.t));
        fireProgress("Sorting LAS points", 100, 100);

        double minTime = points.get(0).t;
        double maxTime = points.get(points.size() - 1).t;

        if (minTime >= maxTime) {
            throw new IOException("ALS file contains inconsistent time information, minimum time " + minTime + " >= maximum time " + maxTime);
        }

        return points;
    }

    private List<TrajectoryPoint> readTrajectory(final CSVFile file) throws Exception {

        LOGGER.info("Reading trajectory file " + file.getName());
        List<TrajectoryPoint> points = new ArrayList();

        // number of trajectory points
        int nPoint = FileManager.getLineNumber(file.getAbsolutePath());
        int count = 0;

        BufferedReader reader = new BufferedReader(new FileReader(file));
        // skip header
        if (file.containsHeader()) {
            reader.readLine();
            count++;
        }

        // skip additional lines
        for (long l = 0; l < file.getNbOfLinesToSkip(); l++) {
            reader.readLine();
            count++;
        }

        // assign column index
        Map<String, Integer> columnAssignment = file.getColumnAssignment();
        // time column
        Integer timeIndex = columnAssignment.get("Time");
        if (timeIndex == null) {
            timeIndex = 3;
        }
        // easting column
        Integer eastingIndex = columnAssignment.get("Easting");
        if (eastingIndex == null) {
            eastingIndex = 0;
        }
        // northing column
        Integer northingIndex = columnAssignment.get("Northing");
        if (northingIndex == null) {
            northingIndex = 1;
        }
        // elevation column
        Integer elevationIndex = columnAssignment.get("Elevation");
        if (elevationIndex == null) {
            elevationIndex = 2;
        }

        String line;
        // parse lines
        while ((line = reader.readLine()) != null) {
            fireProgress("Reading trajectory file " + file.getName(), count++, nPoint);
            String[] lineSplit = line.split(file.getColumnSeparator());
            points.add(new TrajectoryPoint(
                    Double.valueOf(lineSplit[eastingIndex]),
                    Double.valueOf(lineSplit[northingIndex]),
                    Double.valueOf(lineSplit[elevationIndex]),
                    Double.valueOf(lineSplit[timeIndex])));
        }

        LOGGER.info("Sorting and trimming trajectory points");
        fireProgress("Sorting trajectory points", 50, 100);
        // sort points by time stamp
        Collections.sort(points, (TrajectoryPoint o1, TrajectoryPoint o2) -> Double.compare(o1.t, o2.t));
        fireProgress("Sorting trajectory points", 100, 100);

        return points;
    }

    private AlsShot nextShot(final int pointIndex, final int shotIndex) throws InconsistentAlsException {

        List<LasPoint> points = nextLasPoints(pointIndex);

        if (points != null) {
            // check shot alignement
            checkCollinearity(points, EPSILON);
            // check echo consistency (unique rank and same number of echoes)
            checkEchoConsistency(points);

            // take any point from the shot, apply transformation
            LasPoint point = points.get(0);
            Vec4D point_t = Mat4D.multiply(vopMatrix, new Vec4D(point.x, point.y, point.z, 1));
            // shot origin, apply transformation
            Point3d origin = findOrigin(point);
            Vec4D origin_t = Mat4D.multiply(vopMatrix, new Vec4D(origin.x, origin.y, origin.z, 1));
            origin = new Point3d(origin_t.x, origin_t.y, origin_t.z);

            // shot direction
            Vec3D dir3D = Vec3D.normalize(Vec3D.createVec3DFromPoints(
                    new Point3D(origin_t.x, origin_t.y, origin_t.z),
                    new Point3D(point_t.x, point_t.y, point_t.z)));
            Vector3d direction = new Vector3d(dir3D.x, dir3D.y, dir3D.z);
            // number of echoes
            int nEcho = point.n;
            // echoes attributes
            double[] ranges = new double[nEcho];
            int[] classifications = new int[nEcho];
            float[] intensities = new float[nEcho];
            for (LasPoint pt : points) {
                // range (distance from source)
                Vec4D pt_t = Mat4D.multiply(vopMatrix, new Vec4D(pt.x, pt.y, pt.z, 1));
                ranges[pt.r - 1] = distance(origin_t, pt_t);
                // classification
                classifications[pt.r - 1] = pt.classification;
                intensities[pt.r - 1] = (pt.classification == LasPoint.CLASSIFICATION_GROUND)
                        ? (int) (pt.i * RATIO_REFLECTANCE_VEGETATION_SOL)
                        : pt.i;
            }
            // create new ALS shot
            AlsShot alsShot = new AlsShot(
                    shotIndex,
                    origin, direction,
                    ranges, classifications, intensities);
            alsShot.time = point.t;

            return alsShot;
        }

        return null;
    }

    /**
     * Returns a subset of LAS points with the same GPS time from the set of LAS
     * points, starting from given index.
     *
     * @param fromIndex, the index of the first ALS point of the subset.
     * @return the subset of LAS points that have the same GPS time than LAS
     * point at index {@code fromIndex}.
     */
    private List<LasPoint> nextLasPoints(final int fromIndex) {

        if (fromIndex > lasPoints.size() - 1) {
            return null;
        }

        List<LasPoint> pointsWithSameTime = new ArrayList();
        pointsWithSameTime.add(lasPoints.get(fromIndex));
        double time = pointsWithSameTime.get(0).t;
        for (int iPoint = fromIndex + 1; iPoint < lasPoints.size(); iPoint++) {
            if (lasPoints.get(iPoint).t == time) {
                pointsWithSameTime.add(lasPoints.get(iPoint));
            } else {
                break;
            }
        }
        return pointsWithSameTime;
    }

    /**
     * Check whether the subset of LAS points is consistent in terms of echo
     * rank and number of echoes. Every LAS point should have a unique echo rank
     * and the same number of echoes.
     *
     * @param points, a subset of LAS points supposedly belonging to same shot.
     * @return true if the LAS points have unique echo rank and same total echo
     * number.
     */
    private boolean checkEchoConsistency(final List<LasPoint> points) throws InconsistentAlsException {

        boolean flawed = false;
        List<Integer> ranks = new ArrayList();
        int nEcho = points.get(0).n;
        for (LasPoint point : points) {
            if (ranks.contains(point.r)) {
                flawed = true;
            }
            ranks.add(point.r);
            if (nEcho != point.n) {
                flawed = true;
            }
            nEcho = point.n;
        }

        if (flawed) {
            StringBuilder sb = new StringBuilder();
            sb.append("LAS points with same GPS time (i.e. belonging to same shot) seem to have inconsistent echo ranks or echo numbers:");
            for (LasPoint point : points) {
                sb.append('\n').append("  ").append(point);
            }
            throw new InconsistentAlsException(sb.toString());
        }

        return true;
    }

    /**
     * Checks collinearity of the collection of LAS points by ensuring that
     * every combination of three points is collinear, with margin of error.
     *
     * @param points, a list of LAS points.
     * @param epsilon, deviation from one, one meaning strict collinearity.
     * @return true if all the points are (approximately) collinear.
     */
    private boolean checkCollinearity(final List<LasPoint> points, final double epsilon) throws InconsistentAlsException {

        boolean collinear = true;
        StringBuilder sb = new StringBuilder();
        if (points.size() >= 3) {
            int pmax = points.size();
            for (int p1 = 0; p1 < pmax - 2; p1++) {
                for (int p2 = p1 + 1; p2 < pmax - 1; p2++) {
                    for (int p3 = p2 + 1; p3 < pmax; p3++) {
                        if (!collinearity(points.get(p1), points.get(p2), points.get(p3), epsilon)) {
                            collinear = false;
                            // append error and carry on with collinearity test
                            // in case some others points are not lined up
                            if (sb.length() > 0) {
                                sb.append('\n');
                            }
                            sb.append("LAS points with same GSP time are not in line:\n");
                            sb.append("  ").append(points.get(p1)).append('\n');
                            sb.append("  ").append(points.get(p2)).append('\n');
                            sb.append("  ").append(points.get(p3));
                        }
                    }
                }
            }
        }
        if (!collinear) {
            throw new InconsistentAlsException(sb.toString());
        }

        return true;
    }

    /**
     * Check collinearity of the three points, with margin of error.
     * |AB.AC|/|AB||AC|~1
     *
     * @param lpA, first LAS point
     * @param lpB, second LAS point
     * @param lpC, third LAS point
     * @param epsilon, deviation from one, one meaning strict collinearity.
     * @return true if the three LAS points are approximately on same line.
     */
    private boolean collinearity(final LasPoint lpA, final LasPoint lpB, final LasPoint lpC, final double epsilon) {

        Point3D A = new Point3D(lpA.x, lpA.y, lpA.z);
        Point3D B = new Point3D(lpB.x, lpB.y, lpB.z);
        Point3D C = new Point3D(lpC.x, lpC.y, lpC.z);
        Vec3D AB = Vec3D.normalize(Vec3D.createVec3DFromPoints(A, B));
        Vec3D AC = Vec3D.normalize(Vec3D.createVec3DFromPoints(A, C));
        return Math.abs(1.d - Math.abs(Vec3D.dot(AB, AC))) < epsilon;
    }

    /**
     * Given the ALS device trajectory, find out the coordinate of the origin of
     * the current LAS point.
     *
     * @param point, the current LAS point
     * @return the coordinate of the origin of the current LAS point
     */
    private Point3d findOrigin(final LasPoint point) {

        // index of the trajectory point behind current LAS point
        int behindTrjPointIndex = nearestTrajectoryPoint(point.t, Search.MIN);
        // index of the trajectory point ahead of current LAS point
        int aheadTrjPointIndex = behindTrjPointIndex + 1;

        // handle invalid trajectory file
        // should never happen since the trimming of the trajecotry is already
        // handled in PointsToShot.java
        if (behindTrjPointIndex < 0) {
            StringBuilder sb = new StringBuilder();
            sb.append("LAS point GPS time outside trajectory time span.").append('\n');
            sb.append("  ").append(point).append('\n');
            sb.append("  Trajectory time span ")
                    .append(trajectory.get(0).t).append(" ")
                    .append(trajectory.get(trajectory.size() - 1).t);
            throw new IndexOutOfBoundsException(sb.toString());
        }

        // GPS time of the trajectory point behind current LAS point
        double beforeTrjPointTime = trajectory.get(behindTrjPointIndex).t;
        // GPS time of the trajectory point ahead of current LAS point
        double afterTrjPointTime = trajectory.get(aheadTrjPointIndex).t;

        // coordinate of the origin at the time of current LAS point,
        // computed as a linear interpolation between trajectory points
        // before and after current LAS point time.
        Point3d origin = new Point3d();
        double ratio = (afterTrjPointTime != beforeTrjPointTime)
                ? (point.t - beforeTrjPointTime) / (afterTrjPointTime - beforeTrjPointTime)
                : 0.5d;
        // linear interpolation
        origin.x = trajectory.get(behindTrjPointIndex).x + ((trajectory.get(aheadTrjPointIndex).x - trajectory.get(behindTrjPointIndex).x) * ratio);
        origin.y = trajectory.get(behindTrjPointIndex).y + ((trajectory.get(aheadTrjPointIndex).y - trajectory.get(behindTrjPointIndex).y) * ratio);
        origin.z = trajectory.get(behindTrjPointIndex).z + ((trajectory.get(aheadTrjPointIndex).z - trajectory.get(behindTrjPointIndex).z) * ratio);

        return origin;
    }

    @Override
    public IteratorWithException<Shot> iterator() {
        AlsShotIterator it = new AlsShotIterator(lasPoints.size());
        new Thread(it).start();
        return it;
    }

    @Override
    public boolean isCancelled() {
        return cancelled;
    }

    @Override
    public void setCancelled(boolean cancelled) {
        this.cancelled = cancelled;
    }

    private double distance(Vec4D v1, Vec4D v2) {
        return Math.sqrt(Math.pow(v1.x - v2.x, 2) + Math.pow(v1.y - v2.y, 2) + Math.pow(v1.z - v2.z, 2));
    }

    private int nearestTrajectoryPoint(double value, Search search) {

        int low = 0;
        int high = trajectory.size() - 1;

        // special case value smaller than first element of the list
        if (value < trajectory.get(low).t) {
            return (search == Search.MIN) ? -1 : low;
        }

        // special case value bigger than last element of the list
        if (value > trajectory.get(high).t) {
            return (search == Search.MAX) ? -1 : high;
        }

        while (low <= high) {
            int mid = (low + high) >>> 1; // (low + high) / 2
            double midVal = trajectory.get(mid).t;

            if (midVal < value) {
                low = mid + 1;
            } else if (midVal > value) {
                high = mid - 1;
            } else {
                return mid; // key found
            }
        }

        // at this stage we have high < low
        // list[high] < value < list[low]
        return (search == Search.MIN) ? high : low;
    }

    private enum Search {
        MIN, MAX;
    }

    private class AlsShotIterator implements Runnable, IteratorWithException<Shot> {

        private int shotIndex;
        private int pointIndex;
        private final int nPoint;
        private final BlockingQueue<AlsShot> queue = new ArrayBlockingQueue(SHOT_BUFFER);
        private Exception error;

        AlsShotIterator(int nPoint) {
            this.nPoint = nPoint;
        }

        @Override
        public boolean hasNext() throws Exception {
            return pointIndex < nPoint;
        }

        @Override
        public AlsShot next() throws Exception {
            if (null != error) {
                throw error;
            }
            fireProgress("Iterating over LAS points", pointIndex, nPoint);
            return queue.take();
        }

        @Override
        public void run() {

            try {
                AlsShot shot;
                while (null != (shot = nextShot(pointIndex, shotIndex))) {
                    queue.put(shot);
                    shotIndex++;
                    pointIndex += shot.getEchoesNumber();
                }
            } catch (InconsistentAlsException | InterruptedException ex) {
                error = ex;
            }
        }
    }
}
